/* DSP LABORATORY 
 * University of Technology Graz, 2018
 * Josef Kulmer
 * 
 * Fouriertransform calculates the spectrum of a provided input signal. The 
 * input signal is multiplied by a hamming window. The outcome is averaged and
 * forwarded to the output.
 * */

#include <fftw3.h>
#include "math.h"
#include <random>
#include <iostream>
#include "blockprocessing.h"
#include "Fouriertransform.h"

Fouriertransform::Fouriertransform() { //constructor
    // initialize fft operations
    inft = (fftw_complex*) fftw_malloc(sizeof(fftw_complex) * FRAMELENGTH);
    outft = (fftw_complex*) fftw_malloc(sizeof(fftw_complex) * FRAMELENGTH);
    p = fftw_plan_dft_1d(FRAMELENGTH, inft, outft, FFTW_FORWARD, FFTW_ESTIMATE);
    
    // initialize hamming analysis window
    windowfcn.resize(FRAMELENGTH);
    for(unsigned int i=0;i<FRAMELENGTH;i++){
        windowfcn[i]=0.54-0.46*cos(2*PI*i/(FRAMELENGTH-1)); // hamming window equation
    }
    
    // initialize welch periodogram
    welchInternal.resize(FRAMELENGTH);
    welch.resize(FRAMELENGTH);
    iAVERAGING=0;
    
    // setup random numbers
    unif = std::uniform_real_distribution<float>(-1,1);
    filewrite=0;
};

Fouriertransform::~Fouriertransform() { //deconstructor
    fftw_destroy_plan(p);
    fftw_free(outft);
    fftw_free(inft);
};

void Fouriertransform::doPSD(const void *in, const void *out, unsigned long framelength) {
// performs the Fouriertransform
    float *input = (float*) in;
    float *output = (float*) out;
    
    /* multiply input data with window function and copy the result to input variable
     * of the Fourier transform processing */
    for(int i=0;i<framelength;i++){
        inft[i][0] = (double) (*input++ + ADDNOISE*unif(re)) * windowfcn[i];
        inft[i][1] = 0;
    }

    // perform Fourier transform
    fftw_execute(p);
    
    // add present periodogram to saved ones
    for(int i=0;i<framelength;i++){
        welchInternal[i] += (outft[i][0]*outft[i][0] + outft[i][1]*outft[i][1])/AVERAGING;
    }
    
    iAVERAGING+=1; // monitor state of averaging
    
    // if number of averaging is reached, delete internal PSD memory and forward the result
    if (iAVERAGING==AVERAGING){
        iAVERAGING=0;
        for(int i=0;i<framelength;i++){
            welch[i] = welchInternal[i]; // save PSD in welch
            welchInternal[i]=0;
        }
        // write welch PSD to output once
        if (!filewrite) {
            std::cout << std::endl << "Power spectral density estimate" << std::endl;
            for(int i=0; i<framelength;i++){
                std::cout << welch[i] << "," << std::endl;
            }
            filewrite=1;
        }
    }
    
    // provide welch to output
    for(int i=0;i<framelength;i++){
        *output++ = (float) welch[i];
    }
}
